import os
import sys
sys.path.append(os.getcwd())
sys.path.append(os.getcwd()+'/saffronorchestrator')
from pymongo import MongoClient
from searchdatamodels import Candidate, WorkExperience, create_embedding, SearchTemplate
from fastapi.encoders import jsonable_encoder
from rankexpansion import generate_mongo_query_from_template_and_embedding
import unittest
import random
import string
from orchestration import candidate_search_flow
from unittest.mock import patch
import time

class BehaviorTest(unittest.TestCase):
    def setUp(self):
        ATLAS_URI='mongodb+srv://james2:vatVRC5XhsRL6KPV@cluster0.lh8qz.mongodb.net/?retryWrites=true&w=majority'
        client = MongoClient(ATLAS_URI)
        self.collection=client['test_db'].get_collection('candidate_collection')
        self.user_query="banker at goldman sachs"
        
        self.search_template=SearchTemplate(company=["goldman sachs"], title=["banker"])
        self.candidate_list=[
            Candidate(Name=''.join(random.choices(string.ascii_uppercase,k=4)), 
                      WorkExperienceList=[WorkExperience(Specialization="banker", Institution="goldman sachs")],
                      Sources=[''.join(random.choices(string.ascii_uppercase,k=20))]) for _ in range(10)
        ]
        self.embedding=create_embedding(self.user_query)
        n_deleted=self.collection.delete_many({"WorkExperienceList.Specialization":"banker"})
        self.mongo_query=generate_mongo_query_from_template_and_embedding(self.embedding,10,self.search_template)
        time.sleep(3) #mongo can have race conditions with deleting/adding
    
    @patch('orchestration.external_web_sweep')
    def test_database_complete(self,external_web_sweep_mock):
        '''
        GIVEN a user_query, and a mongoDB query_document, and a mongo db collection n_queries=10, and 10 candidates that match the mongodb query
        WHEN the 10 candidates are put into the collection, and we call candidate_search_flow
        THEN we should return the 10 candidates and NOT have called external_web_sweep
            and the candidates should be in the database but with different IDs
        '''
        print('complete')
        #given
        external_web_sweep_mock.side_effect=Exception("should NOT have called external_web_sweep!!!")
        #when
        inserted_id_list=[self.collection.insert_one(jsonable_encoder(c)).inserted_id for c in self.candidate_list ]
        time.sleep(3)
        returned_candidate_list=candidate_search_flow(self.user_query, self.mongo_query, self.collection, 10)
        #then
        returned_candidate_name_set=set([c.Name for c in returned_candidate_list])
        candidate_name_set=set([c.Name for c in self.candidate_list])
        self.assertEqual(len(returned_candidate_name_set), len(candidate_name_set))
        for c_name in candidate_name_set:
            self.assertIn(c_name, returned_candidate_name_set)
            db_id=self.collection.find_one({"Name":c_name})["_id"]
            self.assertNotIn(db_id, inserted_id_list)
        
        #cleanup
        for name in returned_candidate_name_set:
            self.collection.delete_one({"Name":name})
        return
    
    @patch('orchestration.external_web_sweep')
    def test_database_incomplete(self,external_web_sweep_mock):
        '''
        GIVEN an user_query, and a mongoDB query_document, and a collection, n_queries=10, and 9 candidates that match the mongodb query
            and we have set a mock so basilsearchsweeper.external_web_sweep returns [candidate X]
        WHEN the 9 candidates are put into the database, and we call candidate_search_flow
        THEN we should return the 9 candidates + candidate X, and candidate X should be inserted into the collection
        ''' 
        print('incomplete')
        external_web_sweep_mock.return_value=[self.candidate_list[0]]
        #when
        inserted_id_list=[self.collection.insert_one(jsonable_encoder(c)).inserted_id for c in self.candidate_list[1:] ]
        time.sleep(3)
        returned_candidate_list=candidate_search_flow(self.user_query, self.mongo_query, self.collection, 10)
        #then
        returned_candidate_name_set=set([c.Name for c in returned_candidate_list])
        candidate_name_set=set([c.Name for c in self.candidate_list])
        self.assertEqual(len(returned_candidate_name_set), len(candidate_name_set))
        for c_name in candidate_name_set:
            self.assertIn(c_name, returned_candidate_name_set)
            db_id=self.collection.find_one({"Name":c_name})["_id"]
            self.assertNotIn(db_id, inserted_id_list)
        
        #cleanup
        for name in returned_candidate_name_set:
            self.collection.delete_one({"Name":name})
        return
    
    @patch('orchestration.external_web_sweep')
    def test_database_empty(self,external_web_sweep_mock):
        '''
        GIVEN an user_query, and a mongoDB query_document, and a collection, n_queries=10, and 10 candidates
            and we have set a mock so basilsearchsweeper.external_web_sweep returns the 10 candidates
        WHEN we call candidate_search_flow
        THEN we should return the 10 candidates, and they should be inserted into the collection
        ''' 
        external_web_sweep_mock.return_value=self.candidate_list
        #when
        returned_candidate_list=candidate_search_flow(self.user_query, self.mongo_query, self.collection, 10)
        #then
        returned_candidate_name_set=set([c.Name for c in returned_candidate_list])
        candidate_name_set=set([c.Name for c in self.candidate_list])
        self.assertEqual(len(returned_candidate_name_set), len(candidate_name_set))
        for c_name in candidate_name_set:
            self.assertIn(c_name, returned_candidate_name_set)
        
        #cleanup
        for name in returned_candidate_name_set:
            self.collection.delete_one({"Name":name})
        return
    
    @patch('orchestration.external_web_sweep')
    def test_database_incomplete_with_redundancy(self,external_web_sweep_mock):
        '''
        GIVEN an user_query, and a mongoDB query_document, and a collection, n_queries=20, and 10 candidates
            and we have set a mock so basilsearchsweeper.external_web_sweep returns the 10 candidates
        WHEN the 10 candidates have been put into the collection and we call candidate_search_flow
        THEN we should return the 10 candidates (redundant ones should have been removed)
        '''
        #given
        external_web_sweep_mock.return_value=self.candidate_list
        
        #when
        inserted_id_list=[self.collection.insert_one(jsonable_encoder(c)).inserted_id for c in self.candidate_list ]
        time.sleep(3)
        returned_candidate_list=candidate_search_flow(self.user_query, self.mongo_query, self.collection, 20)
        
        #then
        returned_candidate_name_set=set([c.Name for c in returned_candidate_list])
        candidate_name_set=set([c.Name for c in self.candidate_list])
        self.assertEqual(len(returned_candidate_name_set), len(candidate_name_set))
        for c_name in candidate_name_set:
            self.assertIn(c_name, returned_candidate_name_set)
            db_id=self.collection.find_one({"Name":c_name})["_id"]
            self.assertNotIn(db_id, inserted_id_list)
        
        #cleanup
        for name in returned_candidate_name_set:
            self.collection.delete_one({"Name":name})
        return
    
if __name__=='__main__':
    unittest.main()


from searchdatamodels import Candidate
from pymongo import MongoClient
from pymongo.collection import Collection
from basilsweepercrawler import external_web_sweep
from nutmegredundancysolver import merge_redundant_candidates, LSA
from fastapi.encoders import jsonable_encoder
from rankexpansion import rank_candidates

def candidate_search_flow(user_query:str, mongo_query:dict, mongo_collection:Collection, n_queries:int)-> list[Candidate]:
    '''The function `candidate_search_flow` takes a user query, a MongoDB query, a MongoDB collection, and
    the number of queries as input, and returns a list of ranked candidates based on the user query.
    
    Parameters
    ----------
    user_query : str
        The user's query, which is a string representing the search query entered by the user.
    mongo_query : dict
        The `mongo_query` parameter is a dictionary that specifies the query to be executed on the MongoDB
    collection. It is used to filter the candidates based on certain criteria.
    mongo_collection : Collection
        The `mongo_collection` parameter is an instance of the `Collection` class from the `pymongo`
    library. It represents a collection in a MongoDB database where candidate documents are stored.
    n_queries : int
        The parameter `n_queries` represents the desired number of candidates to be retrieved from the
    MongoDB collection. If the number of candidates retrieved from the collection is less than
    `n_queries`, additional candidates will be obtained through an external web sweep.
    
    Returns
    -------
        The function `candidate_search_flow` returns a list of ranked candidates.
    
    '''
    cursor=mongo_collection.aggregate([mongo_query])
    raw_mongo_candidate_list=list(cursor)
    old_id_list=[raw_mongo_candidate["_id"] for raw_mongo_candidate in raw_mongo_candidate_list]
    for raw_mongo_candidate in raw_mongo_candidate_list:
        del raw_mongo_candidate["_id"]
    candidate_list=[Candidate(**raw_mongo_candidate) for raw_mongo_candidate in raw_mongo_candidate_list]
    print("len before sweeper",len(candidate_list))
    if len(candidate_list) < n_queries:
        candidate_list+=external_web_sweep(user_query)
    print("len before redundnacy", len(candidate_list))
    candidate_list=merge_redundant_candidates(candidate_list,LSA)
    print("len after redundnacy", len(candidate_list))
    mongo_collection.insert_many([jsonable_encoder(cand) for cand in candidate_list])
    delete_count=0
    for old_id in old_id_list:
        delete_count+=mongo_collection.delete_many({"_id": old_id}).deleted_count
    print('delete_count ', delete_count)
    return rank_candidates(user_query, candidate_list)
    